/*
 * EXE Info - a tool to extract information from executable files
 *
 * Copyright (C) 2008  Denver Gingerich
 *
 * The contents of this file can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see http://www.gnu.org/licenses/.
 *
 */

#include <stdio.h>	/* printf(), f*() */
#include <errno.h>	/* errno */
#include <string.h>	/* strerror() */
#include <sys/stat.h>	/* stat(), struct stat */
#include <stdlib.h>	/* malloc(), free() */
#include "bfd.h"	/* bfd* */
#include "md5.h"	/* md5_* */

typedef struct {
	md5_byte_t* data;
	size_t len;
} file_t;

/* contents of dwSignature, which precedes the version number;
 * suggested by Simon at http://ossguy.com/?p=28#comment-674 */
unsigned char dwSignature[] = {
	0xbd, 0x04, 0xef, 0xfe
};

/* modified version of memmem() by Bob Stout, released into the public domain;
 * original in memmem.c found in http://c.snippets.org/code/temp/snip-c.zip */
void *memmem(const void *buf, size_t buflen, const void *pattern, size_t len)
{
      size_t i, j;
      char *bf = (char *)buf, *pt = (char *)pattern;

      if (len > buflen)
            return (void *)NULL;

      for (i = 0; i <= (buflen - len); ++i)
      {
            for (j = 0; j < len; ++j)
            {
                  if (pt[j] != bf[i + j])
                        break;
            }
            if (j == len)
                  return (bf + i);
      }
      return NULL;
}

static void section_print(bfd* abfd, asection* section, void* file)
{
	unsigned int opb = bfd_octets_per_byte (abfd);
	md5_state_t md5state;
	md5_byte_t digest[16];

	unsigned long len;
	unsigned long pos;

	len = (unsigned long) bfd_section_size (abfd, section) / opb;
	pos = (unsigned long) section->filepos;

	md5_init(&md5state);
	md5_append(&md5state, &(((file_t*)file)->data[pos]), len);
	md5_finish(&md5state, digest);

	printf ("\t%2d  %-13s  %08lx  %08lx  ", section->index,
        	bfd_get_section_name (abfd, section), pos, len);
	printf("%02x%02x%02x%02x%02x%02x%02x%02x"
		"%02x%02x%02x%02x%02x%02x%02x%02x\n",
		digest[0], digest[1], digest[2], digest[3], 
		digest[4], digest[5], digest[6], digest[7], 
		digest[8], digest[9], digest[10], digest[11], 
		digest[12], digest[13], digest[14], digest[15]);
}

int main(int argc, char** argv)
{
	FILE* handle;
	char* filename;

	struct stat file_stat;
	file_t file;
	size_t bytes_read;
	md5_state_t md5state;
	md5_byte_t digest[16];

	md5_byte_t* ver_pos;
	unsigned int last_ver_idx;
	unsigned short* ver1;
	unsigned short* ver2;
	unsigned short* ver3;
	unsigned short* ver4;

	bfd* abfd;
	char** matching;

	if (2 != argc) {
		printf( "EXE Info v0.1\n"
			"Copyright 2008 Denver Gingerich\n"
			"http://ossguy.com/exe_info/\n"
			"\n"
			"This program is free software; you may redistribute it under the terms of\n"
			"the GNU General Public License.\n"
			"\n"
			"\n"
		);
		printf("Usage: exe_info <file.exe>\n");
		return -1;
	}
	filename = argv[1];

	if (stat(filename, &file_stat)) {
		printf("could not open file '%s': %s\n", filename, (char*)strerror(errno));
		return -2;
	}
	file.len = file_stat.st_size;

	handle = fopen(filename, "rb");
	if (NULL == handle) {
		printf("could not open file '%s': %s\n", filename, (char*)strerror(errno));
		return -3;
	}

	md5_init(&md5state);
	file.data = (md5_byte_t*)malloc(file.len);
	if (NULL == file.data) {
		printf("could not allocate space for file contents: %s\n", (char*)strerror(errno));
		return -4;
	}

	bytes_read = fread(file.data, sizeof(md5_byte_t), file.len, handle);
	fclose(handle);

	if (bytes_read < file.len) {
		printf("could not read file into buffer\n");
		free(file.data); /* clean up */
		return -5;
	}

	printf("General information:\n");
	printf("\tSize:\t%d bytes\n", file.len);

	md5_append(&md5state, file.data, file.len);
	md5_finish(&md5state, digest);

	printf("\tMD5:\t%02x%02x%02x%02x%02x%02x%02x%02x"
		"%02x%02x%02x%02x%02x%02x%02x%02x\n",
		digest[0], digest[1], digest[2], digest[3], 
		digest[4], digest[5], digest[6], digest[7], 
		digest[8], digest[9], digest[10], digest[11], 
		digest[12], digest[13], digest[14], digest[15]);
	printf("\n");

	printf("Versions:\n");
	last_ver_idx = 0;
	for (;;) {
		ver_pos = (md5_byte_t*)memmem(
			&(file.data[last_ver_idx]), file.len - last_ver_idx,
			dwSignature, sizeof(dwSignature));
		if (NULL == ver_pos) {
			break; /* no more dwSignatures found */
		}
		last_ver_idx = (unsigned int)(ver_pos - file.data) + 12;
		if (file.len - last_ver_idx < 4) {
			break; /* not enough space for a version */
		}

		ver1 = (unsigned short*)&(ver_pos[8]);
		ver2 = (unsigned short*)&(ver_pos[10]);
		ver3 = (unsigned short*)&(ver_pos[12]);
		ver4 = (unsigned short*)&(ver_pos[14]);

		printf("\t@ %08x: %hd.%hd.%hd.%hd\n",
			last_ver_idx, *ver2, *ver1, *ver4, *ver3);
	}
	printf("\n");

	abfd = bfd_openr(filename, NULL);
	if (NULL == abfd) {
		printf("could not open file '%s'\n", filename);
		free(file.data); /* clean up */
		return -6;
	}

	printf("Sections:\n");
	if (bfd_check_format_matches (abfd, bfd_object, &matching)) {
		printf("\t #  name           start     length    MD5\n");
		bfd_map_over_sections(abfd, section_print, (void*)&file);
	} else {
		printf("\tno section information for this file; probably not an executable\n");
	}

	/* clean up */
	bfd_close(abfd);
	free(file.data); 

	return 0;
}
